import sys,os
import multiprocessing
import time

####################################################
# abstact layer of game

class Polyhedron:
    """ This is base class of polyhedron. """
    def __init__(self,faces,face_as_edges,face_as_vertices,layout_hint,layout_hint_type):
        """
        Construct an object from 
        a list of the faces, 
        a dictionary which maps each face to edges containg by it, and
        a dictionary which maps each face to vertices containg by it.
        """
        self.face_as_vertices=dict(face_as_vertices)
        vertices=set([])
        for vs in self.face_as_vertices.values():
            vertices=vertices.union(vs)
        self.vertices=frozenset(vertices)
        self.face_as_edges=dict(face_as_edges)
        edges=set([])
        for es in  self.face_as_edges.values():
            edges=edges.union(es)
        self.edges=frozenset(edges)
        self.faces=frozenset(faces)

        self.layout_hint=[[ row[:] for row in chart] for chart in layout_hint]
        self.layout_hint_type=layout_hint_type

    def get_copied_polyhedron(self):
        return Polyhedron(self.get_faces(),self.get_face_as_edges(),self.get_face_as_vertices(),self.get_layout_hint(),self.get_layout_hint_type())

    def get_euler_characteristic_of(self,faces):
        """
        returns the Euler characteristic for the polyhedron restricted to faces.
        """
        all_faces = self.get_faces()
        set_faces = set([])
        set_edges = set([])
        set_vertices = set([])
        for f in faces:
            if f in all_faces:
                set_faces.add(f)
                set_vertices.update(self.get_vertices_contained_by(f))
                set_edges.update(self.get_edges_contained_by(f))
        return len(set_vertices) - len(set_edges) + len(set_faces)

    def get_faces(self):
        """
        return the set of faces.
        """
        return frozenset(self.faces)

    def get_face_as_vertices(self):
        return dict(self.face_as_vertices)

    def get_face_as_edges(self):
        return dict(self.face_as_edges)

    def get_vertices(self):
        """
        return the set of vertices.
        """
        return frozenset(self.vertices)

    def get_edges(self):
        """
        return the set of edges.
        """
        return frozenset(self.edges)


    def exists_common_edge(self,face_1,face_2):
        vs1=self.get_vertices_contained_by(face_1) 
        vs2=self.get_vertices_contained_by(face_2) 
        for vi in vs1:
            if vi in vs2:
                return True
        return False


    def get_vertices_contained_by(self,face):
        """
        return the set of vertices in face.
        """
        return self.face_as_vertices[face]

    def get_edges_contained_by(self,face):
        """
        return the set of vertices in face.
        """
        return self.face_as_edges[face]

    def get_layout_hint(self):
        """
        return layout hint.
        """
        return self.layout_hint

    def get_layout_hint_type(self):
        """
        return layout hint.
        """
        return self.layout_hint_type


class Board:
    def __init__(self,polyhedron,cells_eachplayer=None,n_turn=None):
        if cells_eachplayer==None:
            cells_eachplayer=[[],[]]
        if n_turn==None:
            n_turn=0
        self.polyhedron=polyhedron.get_copied_polyhedron()
        self.cells=[ cells_i[:] for cells_i in cells_eachplayer ]
        self.n_turn=n_turn

    def get_copied_board(self):
        return Board(self.get_polyhedron(),self.cells,self.n_turn)

    def get_polyhedron(self):
        return self.polyhedron.get_copied_polyhedron()

    def get_point(self,player):
        return self.polyhedron.get_euler_characteristic_of(self.get_cells_of(player))

    def get_all_cells(self):
        return self.polyhedron.get_faces()

    def get_empty_cells(self):
        cells=self.get_all_cells()
        return [ cell for cell in cells if self.get_owner_of(cell) < 0 ]

    def n_empty_cells(self):
        n=0
        cells=self.get_all_cells()
        for cell in cells:
            if self.get_owner_of(cell) < 0:
                n += 1
        return n

    def get_an_empty_cell_at(self,n):
        nn=n
        cells=self.get_all_cells()
        for cell in cells:
            if self.get_owner_of(cell) < 0:
                nn -= 1
                if nn < 0:
                    return cell
        return self.get_an_empty_cell_at(n%(n-nn))

    def get_owner_of(self,cell):
        for (i,cells_player_i) in enumerate(self.cells):
            if cell in cells_player_i:
                return i
        return -1

    def get_winner(self):
        if self.get_point(0) > self.get_point(1) :
            return 0
        else:
            return 1

    def append_cell(self,player,cell):
        self.cells[player].append(cell)

    def extend_cell(self,player,cells):
        self.cells[player].extend(cells)

    def set_cells(self,player,cells):
        self.cells[player]=cells[:]

    def get_cells_of(self,player):
        return self.cells[player][:]

    def pop_cell(self,player):
        return self.cells[player].pop()


class Player:
    def play(self,board,timeout):
        return board.get_an_empty_cell_at(0)

    def played(self,board,cell):
        pass

    def end_game(self,board,winner):
        pass

    def start_game(self,board,player_id):
        self.player_id=player_id
        return True

class Outsider:
    def game_start(self,board):
        pass
    def game_end(self,board,winner):
        pass
    def turn_end(self,board):
        pass

class Referee:
    def __init__(self,board,players,timeouts,outsiders=None):
        self.board=board.get_copied_board()
        self.players=tuple(players)
        self.timeouts=timeouts[:]
        if outsiders:
            self.outsiders=outsiders[:]
        else:
            self.outsiders=[]

    def start_game(self):
        self.announce_game_start()

        while self.board.n_empty_cells() > 0:
            c = self.play_turn()
            self.announce_turn_end(c)

        winner=self.board.get_winner()
        self.announce_game_end(winner)

        return self.get_board()

    def get_board(self):
        return self.board.get_copied_board()

    def announce_game_start(self):
        for i,player in enumerate(self.players):
            board=self.board.get_copied_board()
            player.start_game(board,i)

        for outsider in self.outsiders:
            board=self.board.get_copied_board()
            outsider.game_start(board)

    def announce_turn_end(self,cell):
        for i,player in enumerate(self.players):
            board=self.board.get_copied_board()
            player.played(board,cell)

        for outsider in self.outsiders:
            board=self.board.get_copied_board()
            outsider.turn_end(board,cell)


    def announce_game_end(self,winner):
        for i,player in enumerate(self.players):
            board=self.board.get_copied_board()
            player.end_game(board,winner)
        for outsider in self.outsiders:
            board=self.board.get_copied_board()
            outsider.game_end(board,winner)

    def get_timeout(self,player_id):
        return self.timeouts[player_id]

    def play_turn(self):
        timeout=self.get_timeout(self.board.n_turn % 2)
        player=self.players[self.board.n_turn % 2]
        board=self.board.get_copied_board()
        cell=self.try_calc_in(timeout,lambda:player.play(board,timeout))
        if self.board.get_owner_of(cell) >= 0:
            cell=self.board.get_an_empty_cell_at(0)
        self.board.append_cell(self.board.n_turn % 2,cell)
        self.board.n_turn += 1
        return cell

    def try_calc_in(self,timeout,func):
        if timeout > 0:
            q = multiprocessing.Queue()
            p = multiprocessing.Process(target= (lambda:q.put(func())) )
            p.start()
            p.join(timeout)
            if p.is_alive():
                p.terminate()
                p.join()
            if q.empty():
                return None
            else:
                return q.get()
        else:
            return func()


#################################################
# Rough implementation of polyhedron, player, logger
class SmallPolyhedron(Polyhedron):
    def __init__(self):
        n_row=4
        n_col=4
        self.n_row=n_row
        self.n_col=n_col

        self.faces=[ (i,j) for i in range(n_row) for j in range(n_col) ]
        self.faces.append((n_row,0))

        rep={}
        for i in range(n_row+1):
            for j in range(n_col+1):
                rep[(i,j)] = self.get_equivalent_cell((i,j))

        self.edges=[]
        for i in range(n_row):
            for j in range(n_col):
                e=[rep[(i,j)], rep[(i,j+1)]]
                self.edges.append(frozenset(e))
                e=[rep[(i,j)], rep[(i+1,j)]]
                self.edges.append(frozenset(e))
                e=[rep[(i,j)], rep[(i+1,j+1)]]
                self.edges.append(frozenset(e))
        e=[rep[(n_row-1,0)], rep[(0,n_col-1)],None]
        self.edges.append(frozenset(e))

        self.vertices=[]
        for i in range(n_row):
            for j in range(n_col):
                v=[rep[(i,j)], rep[(i,j+1)], rep[(i+1,j+1)]]
                self.vertices.append(frozenset(v))
                v=[rep[(i,j)], rep[(i+1,j)], rep[(i+1,j+1)]]
                self.vertices.append(frozenset(v))
        v=[rep[(n_row-1,0)], rep[(0,n_col-1)], rep[(n_row,0)], None]
        self.vertices.append(frozenset(v))

        self.face_as_edges={} 
        self.face_as_vertices={}
        for f in self.faces:
            es = [ e for e in self.edges if f in e ]
            self.face_as_edges[f]=frozenset(es)
            vs = [ v for v in self.vertices if f in v ]
            self.face_as_vertices[f]=frozenset(vs)
        self.edges=frozenset(self.edges)
        self.vertices=frozenset(self.vertices)

        chart=[[self.get_equivalent_cell((i,j)) for j in range(n_col+1)] for i in range(n_row+1)]
        self.layout_hint=[chart]
        self.layout_hint_type="multihex"

    def get_equivalent_cell(self,cell):
        (i,j)=cell
        if i<self.n_row:
            if j<self.n_col:
                return (i,j)
            else:
                return (self.n_row-i,0)
        else:
            if j==0:
                return (i,0)
            else:
                return (0,self.n_col-j)


class NoBrainPlayer(Player):
    def play(self,board,timeout):
        return board.get_an_empty_cell_at(int(time.strftime("%S")))

class SimpleLogger(Outsider):
    def __init__(self,log_prompt=":"):
        self.log_prompt=log_prompt
    def print_log(self,s):
        print(self.log_prompt+s)
    def get_player_string(self,p):
        if p==0:
            return "x"
        else:
            return "o"


    def turn_end(self,board,cell):
        log = self.get_player_string(board.get_owner_of(cell))
        log += ":plays:"
        log += str(cell)
        self.print_log(log)
    def game_end(self,board,winner):
        log = self.get_player_string(winner)
        log += ":wins:"
        self.print_log(log)
        for i in range(2):
            log = self.get_player_string(i)
            log += ":finalscore:"
            log += str(board.get_point(i))
            self.print_log(log)


#################################################
# Plugin loader
class PluginLoader:
    def __init__(self,plugin_dir,relative_path=True):
        if relative_path:
            parent_path = os.path.dirname(os.path.abspath(__file__))
        else:
            parent_path = ""
        plugin_path = os.path.join(parent_path,plugin_dir)
        sys.path.append(plugin_path)
        plugins=[]
        for item in os.listdir(plugin_path):
            if os.path.isdir(os.path.join(plugin_path,item)):
                try:
                    plugins.append(__import__(item))
                except:
                    pass
        self.plugins=plugins

    def get_plugin_list(self):
        return self.plugins

    def get_plugin_name_list(self,key="name"):
        l=[]
        for plugin in self.plugins:
            if 'get_description' in dir(plugin):
                desc=plugin.get_description()
                if key in desc:
                    l.append(desc[key])
                else:
                    l.append("")
            else:
                l.append("")
        return l

#################################################
# Frontend of the game

class ConsoleFrontend:
    def __init__(self):
        self.comment_prompt="## "
        self.players=[HumanPlayerWithConsole(),NoBrainPlayer()]
        self.timeouts=[0,0]
        self.polyhedron=SmallPolyhedron()
        self.new_board=Board
        self.outsiders=[SimpleLogger()]

        self.ai_plugin_loader=PluginLoader('ai_engines')
        self.polyhedron_plugin_loader=PluginLoader('polyhedrons')
        self.board_plugin_loader=PluginLoader('board_factories')
        self.timeout_option=[('no limit',0)]+[(str(i*30)+'S',i*30) for i in range(1,10)]

    def select_player_and_timeout(self):
        players=[]
        timeouts=[]

        for i in range(2):
            message="Select the player "+str(i)

            r=self.select_from_plugin(self.ai_plugin_loader,message,"Select option",label_default="Human")
            if r:
                players.append(r)
                message="Select timelimit for each turn"
                r=self.select_options(message,[l for l,t in self.timeout_option])
                timeouts.append(self.timeout_option[r][1])
            else:
                players.append(HumanPlayerWithConsole())
                timeouts.append(0)
        self.players=players
        self.timeouts=timeouts


    def select_polyhedron(self):
        polyhedron=self.select_from_plugin(self.polyhedron_plugin_loader,"Select a polyhedron","Select option",label_default=None)
        if polyhedron:
            self.polyhedron=polyhedron
        else:
            self.polyhedron=SmallPolyhedron()

    def select_rule(self):
        r=self.select_from_plugin(self.board_plugin_loader,"Select rule","Select option",label_default="Normal rule")
        if r:
            self.new_board=r
        else:
            self.new_board=Board

    def start_game(self):
        board=self.new_board(self.polyhedron)
        referee=Referee(board,self.players,self.timeouts,self.outsiders)
        referee.start_game()

    def print_comment(self,message):
        print(self.comment_prompt+message)

    def select_from_plugin(self,pl,message,message_option,label_default=None):
        plugins=pl.get_plugin_list()
        label=pl.get_plugin_name_list()
        if len(label)==0:
            return None

        if label_default:
            label=[label_default]+label
            r=self.select_options(message,label)-1
        else:
            r=self.select_options(message,label)

        if r<0:
            return None
        p=plugins[r]
        optlist=p.get_option()
        if optlist:
            option_label=[lab for (lab,opt) in optlist]
            r=self.select_options(message_option,option_label)
            opt=optlist[r][1]
        else:
            opt=None
        return p.get_instance(opt)



    def select_options(self,message,label_list):
        for i,label in enumerate(label_list):
            self.print_comment(str(i)+": "+label)
        self.print_comment(" >> "+message+"[Default=0].")

        rc = sys.stdin.readline()
        try:
            i = int(rc)
        except:
            i = 0
        if 0 <= i and i <len(label_list):
            return i
        else:
            return 0


class HumanPlayerWithConsole(Player):
    def __init__(self):
        self.comment_prompt="#"

    def print_comment(self,s):
        print(self.comment_prompt+s)

    def play(self,board,timeout):
        candidate=board.get_empty_cells()
        select_key_dict=dict([(fi,str(i)) for i,fi in enumerate(candidate)])

        self.print_comment("--------------------------------------------------------")
        for chart in self.get_string_of_board(board):
            for li in chart:
                self.print_comment(" "+li)
        self.print_comment(" Turn of "+self.get_player_string(self.player_id))
        self.print_comment(" >> Select the facet [Default=0].")

        rc = sys.stdin.readline()
        try:
            i = int(rc)
        except:
            i = 0
        if 0 <= i and i <len(candidate):
            cell=candidate[i]
        else:
            cell=candidate[0]
        return cell

    
    def start_game(self,board,player_id):
        self.player_id=player_id
        if board.get_polyhedron().get_layout_hint_type() == "multihex":
            return True
        else:
            return False

    def end_game(self,board,winner):
        self.print_comment("--------------------------------------------------------")
        for chart in self.get_string_of_board(board):
            for li in chart:
                self.print_comment(" "+li)
        if self.player_id == winner:
            self.print_comment(" You, the player "+self.get_player_string(self.player_id) +", win.")
        else:
            self.print_comment(" You, the player "+self.get_player_string(self.player_id) +", loose.")
        self.print_comment("--------------------------------------------------------")
        


    def get_string_of_board(self,board):
        layout_hint=board.polyhedron.get_layout_hint()

        faces=[cij for chart in layout_hint for ci in chart for cij in ci]
        faces.sort()
        face_name_dict=dict([(fi,"") for fi in faces])
        faces=list(set([fi for fi,fj in zip(faces,faces[1:]) if fi == fj]))
        faces.sort()
        for i,fi in enumerate(faces):
            sufix= str(i//26) if i//26 > 0 else ""
            face_name_dict[fi]=chr(ord('A')+i%26)+sufix

        r=[]
        for chart in layout_hint:
            r.append(self.get_board_string_from_chart(chart,board,face_name_dict))
        return r

    def get_player_string(self,p):
        if p==0:
            return "x"
        else:
            return "o"

    def get_face_string(self,board,face,face_name_dict,select_key_dict,min_size=3):
        if face:
            o=board.get_owner_of(face)
            if o < 0:
                r=select_key_dict[face]
            else:
                r=self.get_player_string(o)
            if face_name_dict[face]:
                r+="("+face_name_dict[face]+")"
            tl=len(r)
            ll=((min_size-tl)//2)
            rl=min_size-ll-tl
            return ll*" "+r+rl*" "
        else:
            return min_size*" "
    def get_board_string_from_chart(self,chart,board,face_name_dict):
        l=max([len(ci) for ci in chart])
        ch=[ ci[:]+(l-len(ci))*[None] for ci in chart ]

        nullline=[l*[None]]
        max_len=0

        select_key_dict=dict([(fi,str(i)) for i,fi in enumerate(board.get_empty_cells())])

        max_len=max([len(self.get_face_string(board,cij,face_name_dict,select_key_dict)) for ci in ch for cij in ci])
        llen=(max_len-1)//2
        rlen=max_len-1-llen
        facestring=[[ self.get_face_string(board,cij,face_name_dict,select_key_dict,min_size=max_len) for cij in ci] for ci in ch]
        rows=[]

        barI=[[ True if cij or cik else False for cij,cik in zip([None]+ci,ci+[None])] for ci in ch]
        barJ=[[ True if cik or cjk else False for cik,cjk in zip(ci,cj)] for ci,cj in zip(nullline+ch,ch+nullline)]
        barL=[[ True if cik or cjk else False for cik,cjk in zip(ci+[None],[None]+cj)] for ci,cj in zip(nullline+ch,ch+nullline)]
        xtops=[[ "+" if bij else "-" if cij else " "for bij,cij in zip(bi+[None],ci+[None])] for bi,ci in zip(barI+nullline,nullline+ch)]
        xbots=[[ "+" if bij else "-" if cij else " "for bij,cij in zip(bi+[None],[None]+ci)] for bi,ci in zip(nullline+barI,ch+nullline)]
        barIs=[[ "|" if bij else " " for bij in bi] for bi in barI]
        barLs=[[ llen*"-" if bij else llen*" " for bij in bi] for bi in barL]
        barJs=[[ rlen*"-" if bij else rlen*" " for bij in bi]+[""] for bi in barJ]

        hrs=[]
        for i in range(len(ch)+1):
            rowi=""
            for j in range(l+1):
                rowi+=xbots[i][j]
                rowi+=barLs[i][j]
                rowi+=xtops[i][j]
                rowi+=barJs[i][j]
            hrs.append(rowi)
        rows=[]

        for bi,fi in zip(barI,facestring):
            bars=[ "|" if bij else " " for bij in bi]
            rowi=""
            for bj,fij in zip(bars,fi+[""]):
                rowi+=bj+fij
            rows.append(rowi)
            
        lines=[]
        indent=""
        lines.append(hrs[0])
        indent+=(1+llen)*" "
        for hi,li in zip(hrs[1:],rows):
            lines.append(indent+li)
            lines.append(indent+hi)
            indent+=(1+llen)*" "
    
        flag=True

        while flag:
            for li in lines:
                if not li.startswith(" "):
                    flag=False
            if flag:
                for i,li in enumerate(lines):
                    lines[i]=li[1:]
        return lines


def main_console():
    c=ConsoleFrontend()
    while True:
        c.select_player_and_timeout()
        c.select_polyhedron()
        c.select_rule()
        c.start_game()
        break

if __name__ == "__main__":
    main_console()


